%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the method for computing the Imbalance Ratio formulation presented in the paper Over-sampling the minority class in the feature space published in the IEEE Transactions on Neural Networks and Learning Systems.
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/efso
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [newKernelMatrix, newTrainTargets, sample] = applySMOTE(featureSpace,trainTargets,n)
% Apply the SMOTE algorithm in the empirical feature space
% Input variables: 
% *featureSpace: training patterns in the empirical feature space
% *trainTargets: training targets
% *n: If n is a number it represents the k value for the k-nn algorithm in the SMOTE method. If n is a matrix it represents the set of patterns to over-sample (two pair of patterns per new synthetic example)
% Output variables
% *newKernelMatrix: Reconstructed train kernel matrix, including information from synthetic data.
% *newTrainTargets: Vector of final targets, incorporating the label of new synthetic data.
% *sample: New patterns in the empirical feature space.        
        uniqueTargets = unique(trainTargets);
        nOfPattPerClass = sum(repmat(trainTargets,1,size(uniqueTargets,1))==repmat(uniqueTargets',size(trainTargets,1),1));
            
        [minValue,indexMin] = min(nOfPattPerClass);
        [maxValue,indexMax] = max(nOfPattPerClass);
        
	% Depending on n, we will choose the standard SMOTE algorithm or the one that receives the set of patterns to be over-sampled.
        if numel(n) == 1,
            sample = SMOTE((featureSpace(trainTargets==uniqueTargets(indexMin),:))',maxValue-minValue,n);
        else
            sample = probSMOTE((featureSpace(trainTargets==uniqueTargets(indexMin),:))',maxValue-minValue,n);
        end

	% Reconstruct the kernel matrix
        k1 = featureSpace * sample;
        k2 = sample' * sample;
        k0 = featureSpace * featureSpace';
        newKernelMatrix = [k0, k1; k1', k2];
        sample = sample';
        newTrainTargets = [trainTargets; uniqueTargets(indexMin)*ones(maxValue-minValue,1)];
	% Avoid singularity problems
        newKernelMatrix = newKernelMatrix+eye(size(newKernelMatrix,1))*10^-8;
            
end

