%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the method for computing the Imbalance Ratio formulation presented in the paper Over-sampling the minority class in the feature space published in the IEEE Transactions on Neural Networks and Learning Systems.
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/efso
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [align, derivativeValue] = alignmentDerivative(param, trainpatterns, k_ic)
% Compute the kernel-target alignment derivative using the spherical Gaussian kernel
% Input variables: 
% *param: initial parameters for the kernel
% *trainpatterns: set of train patterns
% *k_ic: centred ideal kernel matrix
% Output variables
% *align: alignment value.
% *derivativeValue: derivative value for param.

    sigma = 10^param;
    % Compute kernel matrix
    k = computeKernelMatrix(trainpatterns', trainpatterns',sigma);
    meank = (eye(size(k,1)) - (ones(size(k,1),1)*ones(1,size(k,1)))/size(k,1));
    kd = computeKernelDerivative(trainpatterns', trainpatterns', sigma);

    % Compute different matrix operations needed
    k_c = centerkernel(k);
    n2k_c = sqrt(sum(sum((k_c).*k_c)));
    n2k_ic = sqrt(sum(sum(k_ic.*k_ic)));
    k_ck_ic = (sum(sum(k_c.*k_ic)));

    % Compute the alignment value and derivative
    derivativeValue = -(1/(n2k_ic)) .* ( ((sum(sum(k_ic.*kd)))/n2k_c) - ((k_ck_ic.*(sum(sum(k_c.*kd))))/(n2k_c^(3))));
    align = -(k_ck_ic/(n2k_c .* n2k_ic));

end

function [kernelMatrix] = computeKernelDerivative(patterns1, patterns2, kernelParam)
        Nf1             = size(patterns1, 2);
        Nf2             = size(patterns2, 2);
        kernelMatrix   = zeros(Nf1, Nf2);

            for i = 1:Nf2,
                kernelMatrix(:,i)  = ((sum((patterns1-patterns2(:,i)*ones(1,Nf1)).^2))'/kernelParam^3) .* exp(-sum((patterns1-patterns2(:,i)*ones(1,Nf1)).^2)'/(2*kernelParam^2));
            end

end
