%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the method for computing the Imbalance Ratio formulation presented in the paper Over-sampling the minority class in the feature space published in the IEEE Transactions on Neural Networks and Learning Systems.
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/efso
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [newKernelMatrixTrain, newTrainTargets, trainEFS, newKernelMatrixTest, testEFS] = preferentialOversampling(kernelMatrixTrain, train, kernelMatrixTest, param)
% Perform preferential over-sampling in the EFS
% Input variables: 
% *kernelMatrixTrain: kernel matrix for training.
% *train: struct variable that contains the information about the training set (patterns and targets).
% *kernelMatrixTest: kernel matrix for test.
% *param: struct of parameters for the algorithm.
%	-param.C: SVC parameter for the cost-sensitive SVC used for computing the slack variables. 
%	-param.k: Kernel parameter for the cost-sensitive SVC.
%	-param.n: k in the knn method used for the over-sampling.
%	-param.pDimensions: percentage of dimensions to be selected for the EFS.
% 	-param.b: represents the actual value (if it is a scalar) for the beta parameter for preferential over-sampling, or the set of beta values to test using kernel learning techniques.
% Output variables
% *newKernelMatrixTrain: Reconstructed train kernel matrix, including information from synthetic data.
% *newTrainTargets: Vector of final targets, incorporating the label of new synthetic data.
% *trainEFS: Train patterns in the empirical feature space.
% *newkernelMatrixTest: Reconstructed test kernel matrix, including information from synthetic data.
% *testEFS: Test patterns in the empirical feature space.
    
    % Compute the optimal hyperplane using a cost-sensitive SVM approach
    alg = CSVC_1v1;

    actual.C = param.C;
    actual.k = param.k;
    [model] = alg.train(train, actual);
    [projected] = alg.test(train, model);
    % Extract slack variables for the minority class patterns
    slacks = compute_projected_minority_class(projected,train.targets);
    results = 1./(1+exp(-1.4*slacks));

    % If param.b is a vector, we optimise the convex combinations of the
    % different kernels induced
	if numel(param.b)>1,
		b = param.b;
		uniqueTargets = unique(train.targets);
		nOfPattPerClass = sum(repmat(train.targets,1,size(uniqueTargets,1))==repmat(uniqueTargets',size(train.targets,1),1));
		    
		[minValue,indexMin] = min(nOfPattPerClass);
		[maxValue] = max(nOfPattPerClass);

        	% Compute the empirical feature space
		[featureSpaceTrain featureSpaceTest] = calculateEmpiricalFeatureSpace(kernelMatrixTrain,kernelMatrixTest,param.pDimensions);
		newTrainTargets = [train.targets; uniqueTargets(indexMin)*ones(maxValue-minValue,1)];
		        
		a = zeros(numel(b),1);
       		% Compute ideal kernel matrix
		ideal_grammatrix = zeros(numel(newTrainTargets),numel(newTrainTargets));
		for i=1:numel(newTrainTargets),
		      ideal_grammatrix(i,:) = (newTrainTargets == newTrainTargets(i)); 
		end
		            
		kernelMatrices = zeros(numel(newTrainTargets),numel(newTrainTargets)*numel(b));
		kernelMatricesTest = zeros(numel(newTrainTargets),size(kernelMatrixTest,2)*numel(b));

        	% Store matrices needed for the optimisation
		M = zeros(numel(b),numel(b));

		for i=1:numel(b),
            	% Compute probabilities
		    g = exp(-b(i)*results)/sum(exp(-b(i)*results));
		    nToOversample = maxValue-minValue;
		    
		    probs_patterns = cumsum(g);
            	    % Compute pairs of patterns to over-sample
		    patterns_selected_ov = computePairsOfPatterns(probs_patterns,nToOversample);
		                
            	    % Over-sampling
		    [k{i}, newTrainTargets, sample] = applySMOTE(featureSpaceTrain,train.targets,patterns_selected_ov);
		    kernelMatrices(:,(1+(i-1)*numel(newTrainTargets)):(i*numel(newTrainTargets))) = (centerkernel(k{i}));
		    a(i) = frobeniusInnerProduct(kernelMatrices(:,(1+(i-1)*numel(newTrainTargets)):(i*numel(newTrainTargets))),ideal_grammatrix);
		                
		    k1 = featureSpaceTest*sample';
		    aux = featureSpaceTrain * featureSpaceTest';
		    kernelMatricesTest(:,(1+(i-1)*size(kernelMatrixTest,2)):(i*size(kernelMatrixTest,2))) = [aux; k1'];
		end

		newKernelMatrix = zeros(numel(newTrainTargets),numel(newTrainTargets));
		newkernelMatrixTest = zeros(numel(newTrainTargets),size(kernelMatrixTest,2));
		            
		for i=1:numel(b),
		    for j=1:numel(b),
		        M(i,j) = frobeniusInnerProduct(kernelMatrices(:,(1+(i-1)*numel(newTrainTargets)):(i*numel(newTrainTargets))),  kernelMatrices(:,(1+(j-1)*numel(newTrainTargets)):(j*numel(newTrainTargets))));
		    end
        end
        	% Optimise convex combination
		[x] = quadprog(M,-2*a,[],[],[],[],zeros(1,numel(b)),ones(1,numel(b))*inf,ones(1,numel(b))*(1/numel(b)));
		solutions = x/norm(x,1);
		            
        	% Compute final kernel matrices
		for i=1:numel(b),
		    newKernelMatrixTrain = newKernelMatrix + solutions(i)*k{i};
		    newKernelMatrixTest = newkernelMatrixTest + solutions(i)*kernelMatricesTest(:,(1+(i-1)*size(kernelMatrixTest,2)):(i*size(kernelMatrixTest,2)));
        end
        
        % Recompute EFS
        [trainEFS testEFS] = calculateEmpiricalFeatureSpace(newKernelMatrixTrain,newKernelMatrixTest,param.pDimensions);
    
    % If param.b is a scalar, we use it to compute the over-sampled patterns
    else
        
		uniqueTargets = unique(train.targets);
		nOfPattPerClass = sum(repmat(train.targets,1,size(uniqueTargets,1))==repmat(uniqueTargets',size(train.targets,1),1));
		    
		[minValue,indexMin] = min(nOfPattPerClass);
		[maxValue,indexMax] = max(nOfPattPerClass);

		[featureSpaceTrain featureSpaceTest] = calculateEmpiricalFeatureSpace(kernelMatrixTrain,kernelMatrixTest,param.pDimensions);
		newTrainTargets = [train.targets; uniqueTargets(indexMin)*ones(maxValue-minValue,1)];

		g = exp(-param.b*results)/sum(exp(-param.b*results));
		    % Balanceamos totalmente las clases
		nToOversample = maxValue-minValue;
		        
		probs_patterns = cumsum(g);
		patterns_selected_ov = computePairsOfPatterns(probs_patterns,nToOversample);
		                
		[newKernelMatrixTrain, newTrainTargets, sample] = applySMOTE(featureSpaceTrain,train.targets,patterns_selected_ov);
		                
		k1 = featureSpaceTest*sample';
		aux = featureSpaceTrain * featureSpaceTest';
		newKernelMatrixTest = [aux; k1'];
        	trainEFS = [featureSpaceTrain; sample];
        	testEFS = featureSpaceTest;
        
	end

end

function minoritary_projected = compute_projected_minority_class(projected,targets)

labels = unique(targets);
for i=1:numel(labels),
    n(i) = sum(targets == labels(i)); 
end

[m,index_minoritary_label] = min(n);
minoritary_projected = projected(targets == labels(index_minoritary_label));

end
