%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the method for computing the Imbalance Ratio formulation presented in the paper Over-sampling the minority class in the feature space published in the IEEE Transactions on Neural Networks and Learning Systems.
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/efso
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

classdef CSVC_1v1 < Algorithm
    % Class for the cost-sensitive SVC algorithm
    %   This class derives from the Algorithm Class

    properties
       
        name_parameters = {'C','k'}
        parameters
    end
    
    methods
        
        function obj = SVC_1v1(kernel)
            obj.name = 'Cost-Sensitive Support Vector Machine Classifier with 1vs1 paradigm';
            if(nargin ~= 0)
                 obj.kernelType = kernel;
            else
                obj.kernelType = 'rbf';
            end
            
        end

        function obj = defaultParameters(obj)
            obj.parameters.C = 10.^(-3:1:3);
            obj.parameters.k = 10.^(-3:1:3);
        end

        function [model_information] = runAlgorithm(obj,train, test, parameters)
                param.C = parameters(1);
                param.k = parameters(2);
                
                c1 = clock;
                model = obj.train(train,param);
                c2 = clock;
                model_information.trainTime = etime(c2,c1);
                
                c1 = clock;
                [model_information.projectedTrain,model_information.predictedTrain] = obj.test(train,model);
                [model_information.projectedTest,model_information.predictedTest] = obj.test(test,model);
                c2 = clock;
                model_information.testTime = etime(c2,c1);

                model.algorithm = 'SVC_1v1';
                model.parameters = param;
                model_information.model = model;

        end
        
        function [model]= train( obj, train , param)
            weights = ones(size(train.targets));
            labels = unique(train.targets);
            for i=1:numel(labels),
                n(i) = sum(train.targets == labels(i)); 
            end

            [m,index_minoritary_label] = min(n);
            options = ['-t 2 -c ' num2str(param.C) ' -g ' num2str(param.k) ' -w' num2str(index_minoritary_label) ' ' num2str(max(n)/m) ' -q'];
            model = svmtrain(weights, train.targets, train.patterns, options);

        end
        
        function [projected, testTargets]= test(obj,test, model)
                [testTargets, acc, projected] = svmpredict(test.targets,test.patterns,model, '');
        end      
    end
end
